#! /usr/bin/env python
#
#     Copyright (C) 2005 Ronan Keegan
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Application.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#
# A Python script to take a sequence file in Fasta format and its mtz file
# and extract a bunch of useful information from them for molecular replacement
# and refinement.
#
# Ronan Keegan 1/11/04

import string, os, sys, time
import MTZ_info 
import Nmol_calc


class TargetInfo:
   """ Collect some information about a target structure. """

   def __init__(self):
     self.seqfile = ''
     self.seqfilename = ''
     self.hklin = ''
     self.hklin_filename = ''
     self.pickle_file = ''
     self.chainID = ''
     self.sequence = ''
     self.seqheader = ''
     self.no_of_res = 0
     self.mol_weight = 0.0
     self.resolution = 0.0
     self.no_of_mols = 0 

     self.mtz_coldata=dict()
     self.col_labels=[]
     
     self.Error=[]
     self.Error.append(False)
     self.Error.append('')

     self.AminoAcids = ["A","B","C","D","E","F","G","H","I","K","L","M","N","P","Q","R","S","T","V","W","Y","X","*","Z"]
     self.AminoAcidWeights = [89.09, 132.61, 121.15, 133.10, 147.13, 165.19, 75.07, 155.16, 131.17, 146.19, 131.17, 149.21, 
                              132.12, 115.13, 146.15, 174.20, 105.09, 119.12, 117.15, 204.23, 181.19, 128.16, 0.0, 146.64]
      try:
         self.debug=eval(os.environ['MRBUMP_DEBUG'])
      except:
         self.debug=False

   def printTargetInfo(self):
     """ Print the Target details to the stdout. """

     print "\n"
     print "Sequence file: %s" % self.seqfile
     print "Reflection (MTZ) file: %s" % self.hklin
     print "Number of residues: %d" % self.no_of_res
     print "Molecular Weight (daltons): %.3f" % self.mol_weight
     print "Number of molecules in a.s.u.: %d" % self.no_of_mols
     print "Resolution of collected data (angstroms): %.3f" % self.resolution
     print "\n"


   def setTargetFile(self, seq_file, job_id):
     """ Set the input sequence handler. """
 
     self.seqfile = seq_file
     #self.parseSeqfile(seq_file, job_id)
     self.seqfilename = os.path.split(seq_file)[-1]

   def setTargetMTZdata(self, mtzdata):
     """ Set the input MTZ file handler. """

     self.hklin = mtzdata
     self.hklin_filename = os.path.split(mtzdata)[-1]

   def setPickleFile(self, pickle_file):
      self.pickle_file = pickle_file

   def setTargetChainID(self, job_id):
     """ Set the Chain ID for the target. Derived from the input job name. """

     self.chainID = job_id

   def setTargetSeq_Res(self):
     """ Get the sequence string and the number of residues from the Fasta file. """

     sequence = ''

     handle = open(self.seqfile)
     line = handle.readline()
     
     #Trap for PIR format
     if ">P1" in line or ">p1" in line:
        line = handle.readline()
        line = handle.readline()

     #Trap for Fasta format
     elif ">" in line:
        line = handle.readline()

     while line:
        sequence += ((string.strip(line).replace(" ","")).replace("*","")).upper()
        line = handle.readline()
     handle.close()
     
     self.sequence = sequence
     self.no_of_res = len(sequence)
 
   def setTargetWeight(self):
     """ Work out the molecular weight of the protein from the residues it contains. """

     index = 0
     total = 0
     water_molecular_weight = 18.015

     for AA_id in self.AminoAcids:
       no_of_occurs = string.count(self.sequence, AA_id)
       weight_for_this = no_of_occurs * self.AminoAcidWeights[index]
       total = total + weight_for_this
       index = index + 1
     self.mol_weight = total - ((self.no_of_res-1)*water_molecular_weight)

   def setTargetMTZinfo(self, mrsearchdir, mtz_coldata):
     """ Call the mtzdump wrapper to extract experiment details from the MTZ file. """

     mtz_dump=MTZ_info.Mtzdump()
     mtz_dump.setMTZdumpLogfile(mrsearchdir)
     mtz_dump.setHklin(self.hklin)
     mtz_dump.go()
     time.sleep(2)

     self.resolution = mtz_dump.getResolution()
     self.col_labels=mtz_dump.getColumnData()

     self.mtz_coldata["F"] = mtz_coldata[0] 
     self.mtz_coldata["SIGF"] = mtz_coldata[1] 
     self.mtz_coldata["FREE"] = mtz_coldata[2] 
     
     print "\n"

     # Test to see that the input column labels are actually in the MTZ file and exit if they are not present
     for label in mtz_coldata:
        if label not in self.col_labels:
           print "MTZ file info message: Error: Column label '%s' not found in MTZ file." % label
           print "The column labels found in the MTZ file are:"
           for name in self.col_labels:
              print name
           self.Error[0] = True
           Error_report = "MTZ column label error: column <b>'%s'</b> not found in MTZ<br>\n" % label
           Error_report += "Column labels found in MTZ:<br><br>\n"
           for name in self.col_labels:
              Error_report += "<b>" + name + "</b>, "
           self.Error[1]=Error_report.rstrip(", ")

     #mtz_dump.getColumnData(self.mtz_coldata) 

   def setTargetNum_mols(self):
     """ Work out the number of molecules in the ASU. """

     nmol=Nmol_calc.Nmol_calc()
     self.no_of_mols=nmol.calculate_nmol(self.hklin, self.seqfile)

   def setTargetInfo(self, seq_file, mtzdata, mrsearchdir, mtz_coldata, job_id):
     self.setTargetFile(seq_file, job_id)
     self.setTargetMTZdata(mtzdata)
     self.setTargetChainID(job_id)
     self.setTargetSeq_Res()
     self.parseSeqfile(seq_file, job_id)
     self.setTargetWeight()
     self.setTargetMTZinfo(mrsearchdir,mtz_coldata)
     self.setTargetNum_mols()

     return self.Error

   def parseSeqfile(self, seq_file, job_id):
      """ Parse the sequence file into the correct format so that Molrep can read it properly."""

      # Open the sequence file and read it in to memory
      new_array=[]

      # Set the header for the fasta file
      header='> ' + job_id

      # Create a single string holding the entire sequence
      seq=self.sequence

      # Break up this string into blocks of 80 letters plus the remainder
      # and write it to a new array
      if len(seq) > 80:
         split=len(seq)/80
         remain=len(seq)%80
         for i in range(split):
            new_array.append(seq[i*80:(i+1)*80])
         new_array.append(seq[len(seq)-remain:len(seq)])
      else:
         new_array=seq

      new=string.join(new_array,'')

      if new!=seq:
        print 'Error in sequence parsing: New sequence differs from origional'
	print "New sequence: %s" % new
	print "Old sequence: %s" % seq
        sys.exit(1)

      # Write out the new sequence file

      o=open(seq_file,'w')

      o.write(header + '\n')
      for i in new_array:
         o.write(i + '\n')

      o.close()




